@@ -3,7 +3,7 @@
  */
 
 /*
-Copyright (C) 2004-2016 John E. Davis
+Copyright (C) 2004-2021,2022 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -33,6 +33,11 @@ USA.
  * term(4) man page on an SGI.
  */
 
+/* The ncurses terminfo binary files come in two flavors: A legacy
+ * format that uses 16 bit integers in the number-section, and a new
+ * 32 bit format (ncurses 6.1, from 2018).
+ */
+
 /* Short integers are stored in two 8-bit bytes.  The first byte contains
  * the least significant 8 bits of the value, and the second byte contains
  * the most significant 8 bits.  (Thus, the value represented is
@@ -43,7 +48,7 @@ USA.
  * source and also is to be considered missing.
  */
 
-static int make_integer (unsigned char *buf)
+static int make_integer16 (unsigned char *buf)
 {
    register int lo, hi;
    lo = (int) *buf++; hi = (int) *buf;
@@ -55,6 +60,20 @@ static int make_integer (unsigned char *buf)
    return lo + 256 * hi;
 }
 
+static int make_integer32 (unsigned char *buf)
+{
+   unsigned int u;
+   int i;
+
+   u = (unsigned int)buf[0];
+   u |= ((unsigned int)buf[1])<<8;
+   u |= ((unsigned int)buf[2])<<16;
+   u |= ((unsigned int)buf[3])<<24;
+
+   i = (int)u;
+   return i;
+}
+
 /*
  * The compiled file is created from the source file descriptions of the
  * terminals (see the -I option of infocmp) by using the terminfo compiler,
@@ -62,16 +81,37 @@ static int make_integer (unsigned char *buf)
  * divided into six parts in the following order:  the header, terminal
  * names, boolean flags, numbers, strings, and string table.
  *
- * The header section begins the file.  This section contains six short
- * integers in the format described below.  These integers are (1) the magic
- * number (octal 0432); (2) the size, in bytes, of the names section; (3)
- * the number of bytes in the boolean section; (4) the number of short
- * integers in the numbers section; (5) the number of offsets (short
- * integers) in the strings section; (6) the size, in bytes, of the string
- * table.
+ * The header section begins the file.  This section contains six
+ * short integers in the format described below.  These integers are
+ * (1) the magic number (legacy:0432, 32 bit:01036); (2) the size, in
+ * bytes, of the names section; (3) the number of bytes in the
+ * boolean section; (4) the number of integers in the numbers
+ * section; (5) the number of offsets (short integers) in the strings
+ * section; (6) the size, in bytes, of the string table.
  */
 
-#define MAGIC 0432
+#define MAGIC_LEGACY 0432
+#define MAGIC_32BIT 01036
+
+typedef struct
+{
+   int num_bool;
+   char **bool_caps;		       /* malloced */
+   unsigned char *bool_values;	       /* malloced */
+
+   int num_numeric;
+   char **numeric_caps;		       /* NOT malloced */
+   unsigned char *numeric_values;      /* malloced */
+
+   int num_string;
+   char **string_caps;		       /* NOT malloced */
+   unsigned char *string_offsets;      /* malloced */
+
+   char *heap;			       /* malloced */
+   char *cap_name_heap;		       /* NOT malloced */
+}
+Extended_Cap_Type;
+
 
 /* In this structure, all char * fields are malloced EXCEPT if the
  * structure is SLTERMCAP.  In that case, only terminal_names is malloced
@@ -91,6 +131,8 @@ struct _pSLterminfo_Type
 
    unsigned int num_numbers;
    unsigned char *numbers;
+   unsigned int sizeof_number;
+   int (*make_integer)(unsigned char *);
 
    unsigned int num_string_offsets;
    unsigned char *string_offsets;
@@ -98,6 +140,9 @@ struct _pSLterminfo_Type
    unsigned int string_table_size;
    char *string_table;
 
+   size_t read_offset;		       /* number of bytes currently read */
+
+   Extended_Cap_Type *ext;
 };
 
 static char *tcap_getstr (SLCONST char *, SLterminfo_Type *);
@@ -109,7 +154,11 @@ static FILE *open_terminfo (char *file, SLterminfo_Type *h)
 {
    FILE *fp;
    unsigned char buf[12];
+   int magic;
 
+#ifdef SLANG_UNTIC
+   (void) fprintf (stdout,"# Trying %s\n", file);
+#endif
    /* Alan Cox reported a security problem here if the application using the
     * library is setuid.  So, I need to make sure open the file as a normal
     * user.  Unfortunately, there does not appear to be a portable way of
@@ -122,19 +171,36 @@ static FILE *open_terminfo (char *file, SLterminfo_Type *h)
    fp = fopen (file, "rb");
    if (fp == NULL) return NULL;
 
-   if ((12 == fread ((char *) buf, 1, 12, fp) && (MAGIC == make_integer (buf))))
+   if (12 != fread ((char *)buf, 1, 12, fp))
      {
-	h->name_section_size = make_integer (buf + 2);
-	h->boolean_section_size = make_integer (buf + 4);
-	h->num_numbers = make_integer (buf + 6);
-	h->num_string_offsets = make_integer (buf + 8);
-	h->string_table_size = make_integer (buf + 10);
+	(void) fclose(fp);
+	return NULL;
+     }
+   magic = make_integer16(buf);
+   if (magic == MAGIC_LEGACY)
+     {
+	h->make_integer = make_integer16;
+	h->sizeof_number = 2;
+     }
+   else if (magic == MAGIC_32BIT)
+     {
+	h->make_integer = make_integer32;
+	h->sizeof_number = 4;
      }
    else
      {
-	fclose (fp);
-	fp = NULL;
+	(void) fclose (fp);
+	return NULL;
      }
+
+   h->name_section_size = make_integer16 (buf + 2);
+   h->boolean_section_size = make_integer16 (buf + 4);
+   h->num_numbers = make_integer16 (buf + 6);
+   h->num_string_offsets = make_integer16 (buf + 8);
+   h->string_table_size = make_integer16 (buf + 10);
+
+   h->read_offset = 12;
+
    return fp;
 }
 
@@ -146,7 +212,7 @@ static FILE *open_terminfo (char *file, SLterminfo_Type *h)
  */
 
 /* returns pointer to malloced space */
-static unsigned char *read_terminfo_section (FILE *fp, unsigned int size)
+static void *read_terminfo_section (FILE *fp, SLterminfo_Type *t, unsigned int size)
 {
    char *s;
 
@@ -156,12 +222,14 @@ static unsigned char *read_terminfo_section (FILE *fp, unsigned int size)
 	SLfree (s);
 	return NULL;
      }
+   t->read_offset += size;
+
    return (unsigned char *) s;
 }
 
 static char *read_terminal_names (FILE *fp, SLterminfo_Type *t)
 {
-   return t->terminal_names = (char *) read_terminfo_section (fp, t->name_section_size);
+   return t->terminal_names = (char *) read_terminfo_section (fp, t, t->name_section_size);
 }
 
 /*
@@ -182,18 +250,19 @@ static unsigned char *read_boolean_flags (FILE *fp, SLterminfo_Type *t)
    unsigned int size = (t->name_section_size + t->boolean_section_size) % 2;
    size += t->boolean_section_size;
 
-   return t->boolean_flags = read_terminfo_section (fp, size);
+   return t->boolean_flags = (unsigned char *)read_terminfo_section (fp, t, size);
 }
 
 /*
  * The numbers section is similar to the boolean flags section.  Each
- * capability takes up two bytes, and is stored as a short integer.  If the
- * value represented is -1 or -2, the capability is taken to be missing.
+ * capability takes up 2(4) bytes for the legacy(32 bit) format and
+ * is stored as a integer.  If the value represented is -1 or -2, the
+ * capability is taken to be missing.
  */
 
 static unsigned char *read_numbers (FILE *fp, SLterminfo_Type *t)
 {
-   return t->numbers = read_terminfo_section (fp, 2 * t->num_numbers);
+   return t->numbers = (unsigned char *)read_terminfo_section (fp, t, t->sizeof_number * t->num_numbers);
 }
 
 /* The strings section is also similar.  Each capability is stored as a
@@ -207,7 +276,7 @@ static unsigned char *read_numbers (FILE *fp, SLterminfo_Type *t)
 
 static unsigned char *read_string_offsets (FILE *fp, SLterminfo_Type *t)
 {
-   return t->string_offsets = (unsigned char *) read_terminfo_section (fp, 2 * t->num_string_offsets);
+   return t->string_offsets = (unsigned char *) read_terminfo_section (fp, t, 2 * t->num_string_offsets);
 }
 
 /* The final section is the string table.  It contains all the values of
@@ -217,7 +286,133 @@ static unsigned char *read_string_offsets (FILE *fp, SLterminfo_Type *t)
 
 static char *read_string_table (FILE *fp, SLterminfo_Type *t)
 {
-   return t->string_table = (char *) read_terminfo_section (fp, t->string_table_size);
+   return t->string_table = (char *) read_terminfo_section (fp, t, t->string_table_size);
+}
+
+static void free_ext_caps (Extended_Cap_Type *ext)
+{
+   if (ext == NULL) return;
+
+   SLfree ((char *)ext->bool_values);
+   SLfree ((char *)ext->numeric_values);
+   SLfree ((char *)ext->string_offsets);
+   SLfree ((char *)ext->heap);
+   SLfree ((char *)ext->bool_caps);
+
+   SLfree ((char *)ext);
+}
+
+static int try_read_extended_caps (FILE *fp, SLterminfo_Type *t)
+{
+   Extended_Cap_Type *ext;
+   size_t size;
+   char **cap_names;
+   unsigned char *b;
+   char *heap_max, *cap_name_heap;
+   unsigned int i, num_caps, heap_size;
+   unsigned char buf[11];
+
+   size = 10;
+   b = buf;
+   if (t->read_offset % 2)
+     {
+	size++;
+	b++;
+     }
+
+   if (size != fread (buf, 1, size, fp))
+     return 0;			       /* assume no extended caps */
+   t->read_offset += size;
+
+   if (NULL == (ext = (Extended_Cap_Type *)SLmalloc (sizeof(Extended_Cap_Type))))
+     return -1;
+   memset (ext, 0, sizeof(Extended_Cap_Type));
+
+   ext->num_bool = make_integer16(b);
+   ext->num_numeric = make_integer16(b+2);
+   ext->num_string = make_integer16(b+4);
+   (void) make_integer16(b+6);   /* number of valid strings (not cancelled or absent) */
+   heap_size = make_integer16(b+8);   /* size of the area containing all strings */
+
+   /* The heap should be large enough to hold all the strings */
+   num_caps = ext->num_bool + ext->num_numeric + ext->num_string;
+   if ((ext->num_bool < 0) || (ext->num_numeric < 0) || (ext->num_string < 0)
+       || (heap_size < 2*(num_caps + ext->num_string)))
+     goto return_failure;
+
+   size = ext->num_bool;
+   if (size % 2) size++;     /* so numbers will start on an even byte bndry */
+   if (NULL == (ext->bool_values = (unsigned char *)read_terminfo_section (fp, t, size)))
+     goto return_failure;
+
+   size = t->sizeof_number * ext->num_numeric;
+   if (NULL == (ext->numeric_values = (unsigned char *)read_terminfo_section (fp, t, size)))
+     goto return_failure;
+
+   /* Now read the offsets for the strings.  These also include offsets for all
+    * capability names.
+    */
+   size = 2 * ext->num_string;
+   /* Now add the offsets for _all_ the extended cap names */
+   size += 2 * num_caps;
+   if (NULL == (ext->string_offsets = (unsigned char *)read_terminfo_section (fp, t, size)))
+     goto return_failure;
+
+   /* Now read the heap that contains all the strings. The heaps contains two
+    * areas: the first contains the values the string-valued capabilities.  The
+    * seccond area contains the names of all the extended capabilities.  Unfortunately
+    * the file format lacks explicit information about where the second area begins.
+    * So it will have to be deduced.
+    */
+   if (NULL == (ext->heap = (char *)read_terminfo_section (fp, t, heap_size)))
+     goto return_failure;
+   heap_max = ext->heap + heap_size;
+
+   /* The whole point of this next loop is to find the cap_name_heap.  This info
+    * should have been put in the header.
+    */
+   cap_name_heap = ext->heap;
+   b = ext->string_offsets;
+   for (i = 0; i < (unsigned int)ext->num_string; i++)
+     {
+	int offset = make_integer16 (b);
+	if (((offset >= 0) && ((unsigned int)offset < heap_size))
+	    && (ext->heap + offset >= cap_name_heap))
+	  {
+	     cap_name_heap = ext->heap + offset;
+	     while ((cap_name_heap < heap_max)
+		    && (*cap_name_heap != 0))
+	       cap_name_heap++;
+	     if (cap_name_heap < heap_max) cap_name_heap++;   /* skip \0 */
+	  }
+	b += 2;
+     }
+
+   if (NULL == (cap_names = (char **)SLmalloc(num_caps * sizeof(char *))))
+     goto return_failure;
+
+   for (i = 0; i < num_caps; i++)
+     {
+	int offset = make_integer16 (b);
+
+	if ((offset < 0)
+	    || ((cap_names[i] = cap_name_heap + offset) >= heap_max))
+	  cap_names[i] = "*invalid*";
+
+	b += 2;
+     }
+   ext->cap_name_heap = cap_name_heap;
+
+   ext->bool_caps = cap_names;
+   ext->numeric_caps = cap_names + ext->num_bool;
+   ext->string_caps = ext->numeric_caps + ext->num_numeric;
+
+   t->ext = ext;
+   return 0;
+
+return_failure:
+   free_ext_caps (ext);
+   return 0;			       /* we tried */
 }
 
 /*
@@ -230,10 +425,73 @@ static char *read_string_table (FILE *fp, SLterminfo_Type *t)
  * are implemented by multiple links to the same compiled file.
  */
 
+static FILE *try_open_tidir (SLterminfo_Type *ti, const char *tidir, const char *term)
+{
+   char file[1024];
+
+   if (sizeof (file) > strlen (tidir) + 5 + strlen (term))
+     {
+	FILE *fp;
+
+	sprintf (file, "%s/%c/%s", tidir, *term, term);
+	if (NULL != (fp = open_terminfo (file, ti)))
+	  return fp;
+
+	sprintf (file, "%s/%02x/%s", tidir, (unsigned char)*term, term);
+	if (NULL != (fp = open_terminfo (file, ti)))
+	  return fp;
+     }
+
+   return NULL;
+}
+
+static FILE *try_open_env (SLterminfo_Type *ti, const char *term, const char *envvar)
+{
+   char *tidir;
+
+   if (NULL == (tidir = _pSLsecure_getenv (envvar)))
+     return NULL;
+
+   return try_open_tidir (ti, tidir, term);
+}
+
+static FILE *try_open_home (SLterminfo_Type *ti, const char *term)
+{
+   char home_ti[1024];
+   char *env;
+
+   if (NULL == (env = _pSLsecure_getenv ("HOME")))
+     return NULL;
+
+   strncpy (home_ti, env, sizeof (home_ti) - 11);
+   home_ti [sizeof(home_ti) - 11] = 0;
+   strcat (home_ti, "/.terminfo");
+
+   return try_open_tidir (ti, home_ti, term);
+}
+
+static FILE *try_open_env_path (SLterminfo_Type *ti, const char *term, const char *envvar)
+{
+   char tidir[1024];
+   char *env;
+   unsigned int i;
+
+   if (NULL == (env = _pSLsecure_getenv (envvar)))
+     return NULL;
+
+   i = 0;
+   while (-1 != SLextract_list_element (env, i, ':', tidir, sizeof(tidir)))
+     {
+	FILE *fp = try_open_tidir (ti, tidir, term);
+	if (fp != NULL) return fp;
+	i++;
+     }
+
+   return NULL;
+}
+
 static SLCONST char *Terminfo_Dirs [] =
 {
-   "", /* $TERMINFO */
-   "", /* $HOME/.terminfo */
 #ifdef MISC_TERMINFO_DIRS
    MISC_TERMINFO_DIRS,
 #endif
@@ -248,6 +506,23 @@ static SLCONST char *Terminfo_Dirs [] =
    NULL,
 };
 
+static FILE *try_open_hardcoded (SLterminfo_Type *ti, const char *term)
+{
+   const char *tidir, **tidirs;
+
+   tidirs = Terminfo_Dirs;
+   while (NULL != (tidir = *tidirs++))
+     {
+	FILE *fp;
+
+	if ((*tidir != 0)
+	    && (NULL != (fp = try_open_tidir (ti, tidir, term))))
+	  return fp;
+     }
+
+   return NULL;
+}
+
 void _pSLtt_tifreeent (SLterminfo_Type *t)
 {
    if (t == NULL)
@@ -259,6 +534,7 @@ void _pSLtt_tifreeent (SLterminfo_Type *t)
 	SLfree ((char *)t->string_offsets);
 	SLfree ((char *)t->numbers);
 	SLfree ((char *)t->boolean_flags);
+	free_ext_caps (t->ext);
      }
    SLfree ((char *)t->terminal_names);
    SLfree ((char *)t);
@@ -266,11 +542,7 @@ void _pSLtt_tifreeent (SLterminfo_Type *t)
 
 SLterminfo_Type *_pSLtt_tigetent (SLCONST char *term)
 {
-   SLCONST char **tidirs, *tidir;
    FILE *fp = NULL;
-   char file[1024];
-   static char home_ti [1024];
-   char *env;
    SLterminfo_Type *ti;
 
    if (
@@ -302,36 +574,13 @@ SLterminfo_Type *_pSLtt_tigetent (SLCONST char *term)
    /* If we are on a termcap based system, use termcap */
    if (0 == tcap_getent (term, ti)) return ti;
 
-   if (NULL != (env = _pSLsecure_getenv ("TERMINFO")))
-     Terminfo_Dirs[0] = env;
-
-   if (NULL != (env = _pSLsecure_getenv ("HOME")))
-     {
-	strncpy (home_ti, env, sizeof (home_ti) - 11);
-	home_ti [sizeof(home_ti) - 11] = 0;
-	strcat (home_ti, "/.terminfo");
-	Terminfo_Dirs [1] = home_ti;
-     }
-
-   tidirs = Terminfo_Dirs;
-   while (NULL != (tidir = *tidirs++))
-     {
-	if (*tidir == 0)
-	  continue;
-
-	if (sizeof (file) > strlen (tidir) + 5 + strlen (term))
-	  {
-	     sprintf (file, "%s/%c/%s", tidir, *term, term);
-	     if (NULL != (fp = open_terminfo (file, ti)))
-	       break;
-	     sprintf (file, "%s/%02x/%s", tidir, (unsigned char)*term, term);
-	     if (NULL != (fp = open_terminfo (file, ti)))
-	       break;
-	  }
-     }
+   fp = try_open_env_path (ti, term, "TERMINFO_DIRS");
+   if (fp == NULL) fp = try_open_env (ti, term, "TERMINFO");
+   if (fp == NULL) fp = try_open_home (ti, term);
+   if (fp == NULL) fp = try_open_hardcoded (ti, term);
 
 #ifdef SLANG_UNTIC
-   fp_open_label:
+fp_open_label:
 #endif
 
    if (fp == NULL)
@@ -345,7 +594,8 @@ SLterminfo_Type *_pSLtt_tigetent (SLCONST char *term)
        || (NULL == read_boolean_flags (fp, ti))
        || (NULL == read_numbers (fp, ti))
        || (NULL == read_string_offsets (fp, ti))
-       || (NULL == read_string_table (fp, ti)))
+       || (NULL == read_string_table (fp, ti))
+       || (-1 == try_read_extended_caps (fp, ti)))
      {
 	_pSLtt_tifreeent (ti);
 	ti = NULL;
@@ -377,7 +627,11 @@ static int compute_cap_offset (SLCONST char *cap, SLterminfo_Type *t, Tgetstr_Ma
    char cha, chb;
 
    (void) t;
-   cha = *cap++; chb = *cap;
+
+   /* cap must be at most 2 characters */
+   if ((0 == (cha = cap[0]))
+       || ((0 != (chb = cap[1])) && (0 != cap[2])))
+     return -1;
 
    while (*map->name != 0)
      {
@@ -393,16 +647,36 @@ static int compute_cap_offset (SLCONST char *cap, SLterminfo_Type *t, Tgetstr_Ma
 
 char *_pSLtt_tigetstr (SLterminfo_Type *t, SLCONST char *cap)
 {
-   int offset;
+   int i, offset;
 
    if (t == NULL)
      return NULL;
 
    if (t->flags == SLTERMCAP) return tcap_getstr (cap, t);
 
+   /* Check local extensions first */
+   if (t->ext != NULL)
+     {
+	Extended_Cap_Type *e = t->ext;
+	int n = e->num_string;
+
+	for (i = 0; i < n; i++)
+	  {
+	     char *val;
+	     if (strcmp (cap, e->string_caps[i]))
+	       continue;
+
+	     offset = make_integer16 (e->string_offsets + 2*i);
+	     if (offset < 0) return NULL;
+	     val = e->heap + offset;
+	     if (val >= e->cap_name_heap) return NULL;
+	     return val;
+	  }
+     }
+
    offset = compute_cap_offset (cap, t, Tgetstr_Map, t->num_string_offsets);
    if (offset < 0) return NULL;
-   offset = make_integer (t->string_offsets + 2 * offset);
+   offset = make_integer16 (t->string_offsets + 2 * offset);
    if (offset < 0) return NULL;
    return t->string_table + offset;
 }
@@ -416,9 +690,23 @@ int _pSLtt_tigetnum (SLterminfo_Type *t, SLCONST char *cap)
 
    if (t->flags == SLTERMCAP) return tcap_getnum (cap, t);
 
+   if (t->ext != NULL)
+     {
+	Extended_Cap_Type *e = t->ext;
+	int i, n = e->num_numeric;
+
+	for (i = 0; i < n; i++)
+	  {
+	     if (strcmp (cap, e->numeric_caps[i]))
+	       continue;
+	     return (*t->make_integer)(e->numeric_values + i*t->sizeof_number);
+	  }
+     }
+
    offset = compute_cap_offset (cap, t, Tgetnum_Map, t->num_numbers);
    if (offset < 0) return -1;
-   return make_integer (t->numbers + 2 * offset);
+
+   return (*t->make_integer)(t->numbers + t->sizeof_number * offset);
 }
 
 int _pSLtt_tigetflag (SLterminfo_Type *t, SLCONST char *cap)
@@ -429,6 +717,19 @@ int _pSLtt_tigetflag (SLterminfo_Type *t, SLCONST char *cap)
 
    if (t->flags == SLTERMCAP) return tcap_getflag (cap, t);
 
+   if (t->ext != NULL)
+     {
+	Extended_Cap_Type *e = t->ext;
+	int i, n = e->num_bool;
+
+	for (i = 0; i < n; i++)
+	  {
+	     if (strcmp (cap, e->bool_caps[i]))
+	       continue;
+	     return e->bool_values[i];
+	  }
+     }
+
    offset = compute_cap_offset (cap, t, Tgetflag_Map, t->boolean_section_size);
 
    if (offset < 0) return -1;
@@ -675,7 +976,7 @@ static int tcap_getent (SLCONST char *term, SLterminfo_Type *ti)
 }
 
 /* These routines are provided only for backward binary compatability.
- * They will vanish in V2.x
+ * They will vanish in V3.x
  */
 char *SLtt_tigetent (SLFUTURE_CONST char *s)
 {
@@ -689,7 +990,7 @@ char *SLtt_tigetstr (SLFUTURE_CONST char *s, char **p)
    return _pSLtt_tigetstr ((SLterminfo_Type *) *p, s);
 }
 
-extern int SLtt_tigetnum (SLFUTURE_CONST char *s, char **p)
+int SLtt_tigetnum (SLFUTURE_CONST char *s, char **p)
 {
    if (p == NULL)
      return -1;
