@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2014 John E. Davis
+Copyright (C) 2004-2021,2022 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -76,9 +76,11 @@ static void free_static_sval_token (_pSLang_Token_Type *tok)
 _pSLtok_Type _pSLtoken_init_slstring_token (_pSLang_Token_Type *tok, _pSLtok_Type type,
 					    SLCONST char *s, SLstrlen_Type len)
 {
-   if (NULL == (tok->v.s_val = _pSLstring_make_hashed_string (s, len, &tok->hash)))
+   char *sval;
+   if (NULL == (sval = _pSLstring_make_hashed_string (s, len, &tok->hash)))
      return tok->type = EOF_TOKEN;
 
+   tok->v.s_val = sval;
    tok->free_val_func = free_slstring_token_val;
    return tok->type = type;
 }
@@ -217,12 +219,12 @@ static SLCONST char *map_token_to_string (_pSLang_Token_Type *tok)
 
 #ifdef HAVE_LONG_LONG
       case LLONG_TOKEN:
-	sprintf (numbuf, "%lld", tok->v.llong_val);
+	sprintf (numbuf, SLFMT_LLD, tok->v.llong_val);
 	s = numbuf;
 	break;
 
       case ULLONG_TOKEN:
-	sprintf (numbuf, "%llu", tok->v.ullong_val);
+	sprintf (numbuf, SLFMT_LLU, tok->v.ullong_val);
 	s = numbuf;
 	break;
 #endif
@@ -268,7 +270,7 @@ static SLCONST char *map_token_to_string (_pSLang_Token_Type *tok)
       case FLOAT_TOKEN:
       case DOUBLE_TOKEN:
       case COMPLEX_TOKEN:
-	/* drop */
+	/* fall through */
 #endif
       default:
 	  if (NULL != (s = lookup_op_token_string (type)))
@@ -485,7 +487,7 @@ int _pSLcheck_identifier_syntax (SLCONST char *name)
      }
 
    _pSLang_verror (SL_SYNTAX_ERROR,
-		 "Identifier or structure field name '%s' contains an illegal character", name);
+		 "Identifier, namespace or structure field name '%s' contains an illegal character", name);
    return -1;
 }
 
@@ -541,6 +543,7 @@ static int get_ident_token (_pSLang_Token_Type *tok, unsigned char *s, unsigned
      {
 	tok->v.s_val = table->name;
 	tok->free_val_func = free_static_sval_token;
+	tok->flags |= SLTOKEN_VALUE_IS_RESERVED;
 	return (tok->type = table->type);
      }
 
@@ -841,12 +844,7 @@ static int expand_escaped_string (register char *s,
 	     is_binary = -1;
 	     break;
 	  }
-	if ((isunicode == 0)
-#if 0
-	    && ((wch < 127)
-		|| (utf8_encode == 0))
-#endif
-	    )
+	if (isunicode == 0)
 	  {
 	     if (wch == 0)
 	       is_binary = 1;
@@ -1037,6 +1035,8 @@ static int read_string_token (unsigned char quote_char,
 
 	if (ch == '\\')
 	  {
+	     int cr = 0;
+
 	     if (is_multiline_raw)
 	       {
 		  s[len++] = ch;
@@ -1045,15 +1045,23 @@ static int read_string_token (unsigned char quote_char,
 	       }
 
 	     ch = prep_get_char ();
+	     if (ch == '\r')
+	       {
+		  cr = 1;
+		  ch = prep_get_char ();
+	       }
+
 	     if ((ch == '\n') || (ch == 0))
 	       {
+		  /* ignore the \r, if present */
 		  is_continued = 1;
 		  break;
 	       }
 	     s[len++] = '\\';
 	     if (len < maxlen)
 	       {
-		  s[len++] = ch;
+		  if (cr) s[len++] = '\r';
+		  if (len < maxlen) s[len++] = ch;
 		  has_bs = 1;
 	       }
 	     continue;
@@ -2028,24 +2036,33 @@ void _pSLcompile_byte_compiled (void)
 	   case BREAK_N_TOKEN:
 	   case LINE_NUM_TOKEN:
 	   case CHAR_TOKEN:
-	   case UCHAR_TOKEN:
 	   case SHORT_TOKEN:
-	   case USHORT_TOKEN:
 	   case INT_TOKEN:
-	   case UINT_TOKEN:
 	   case LONG_TOKEN:
+	     if (NULL == check_byte_compiled_token (buf))
+	       return;
+	     tok.v.long_val = SLatol ((unsigned char *)buf);
+	     break;
+
+	   case UCHAR_TOKEN:
+	   case USHORT_TOKEN:
+	   case UINT_TOKEN:
 	   case ULONG_TOKEN:
 	     if (NULL == check_byte_compiled_token (buf))
 	       return;
-	     tok.v.long_val = atol (buf);
+	     tok.v.ulong_val = SLatoul ((unsigned char *)buf);
 	     break;
 #ifdef HAVE_LONG_LONG
 	   case LLONG_TOKEN:
-	   case ULLONG_TOKEN:
 	     if (NULL == check_byte_compiled_token (buf))
 	       return;
 	     tok.v.llong_val = SLatoll ((unsigned char *)buf);
 	     break;
+	   case ULLONG_TOKEN:
+	     if (NULL == check_byte_compiled_token (buf))
+	       return;
+	     tok.v.ullong_val = SLatoull ((unsigned char *)buf);
+	     break;
 #endif
 	   case COMPLEX_TOKEN:
 	   case FLOAT_TOKEN:
@@ -2265,7 +2282,7 @@ static int byte_compile_multiline_token (_pSLang_Token_Type *tok,
 	break;
 
       default:
-	SLang_verror (SL_Internal_Error, "Unsupported multline token: 0x%X", tok->type);
+	SLang_verror (SL_Internal_Error, "Unsupported multiline token: 0x%X", tok->type);
 	return -1;
      }
 
@@ -2345,11 +2362,11 @@ static void byte_compile_token (_pSLang_Token_Type *tok)
 
 #ifdef HAVE_LONG_LONG
       case LLONG_TOKEN:
-	sprintf (b3, "%lld", tok->v.llong_val);
+	sprintf (b3, SLFMT_LLD, tok->v.llong_val);
 	break;
 
       case ULLONG_TOKEN:
-	sprintf (b3, "%llu", tok->v.ullong_val);
+	sprintf (b3, SLFMT_LLU, tok->v.ullong_val);
 	break;
 #endif
       case _BSTRING_TOKEN:
